open Format
open Term
open Substitution
open Overlap
open Rewriting
open Arg
open Result

(* proof tree *)

type proof = 
  | EMPTY  
  | KH12    of Rules.t (* R *) * Rules.t (* S *) * Rules.t (* CP_S(R) *)  * proof
  | KB70    of Rules.t (* R *) * Rules.t (* CP(R) *)
  | HM11    of Rules.t (* R *) * Rules.t (* CPS'(R) *) * Rules.t (* CP(R) *)
  | HM13    of Rules.t (* R *) * Rules.t (* S *) * Rules.t (* CPS'(R,S) *)  * proof * proof
  | HM13CR  of Rules.t (* R *) * Rules.t (* non-closed-CPS(R) *) 
  | HvONO18 of Rules.t (* R *) * Rules.t (* CP(R) *) * Rules.t (* closing steps *) * proof
  | HOT     of Rules.t (* R *) * Rules.t (* C *) * (Rule.t * int) list (* rule label *)
  | OH14b   of Rules.t (* R *) * Rules.t (* CPout(R) *) * Rules.t (* CPin(R) *) * Rules.t (* closing steps *) 
  | H80     of Rules.t (* R *) * Rules.t (* CP(R) *) (* strong closedness *)
  | VO94    of Rules.t (* R *) * Rules.t (* CP(R) *) (* development closedness *)
  | AYT09   of Rules.t (* R *) * Rules.t (* CP(R) *) (* development closedness a la Toyama *)
  | VO08    of Rules.t (* R *) * (Rule.t * int) list (* rule label *)
  | JK86    of Rules.t (* R *) * Rules.t (* CP_AC(R) *)
  | Huet_conjecture

let print_label ppf (rule, n) =
  fprintf ppf "[%a] = %d" Rule.print rule n 

let rec print_proof ppf = function
  | EMPTY ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- R is empty.@."
  | KB70 (rs, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Knuth and Bendix, 1970]@.\
         R:@.%a@.\
         CP(R):@.%a@."
        Rules.print rs
        Rules.print cp
  | HM11 (rs, cps, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Hirokawa and Middeldorp, 2011]@.\
         R:@.%a@.\
         CPS'(R):@.%a@.\
         CP(R):@.%a@."
        Rules.print rs
        Rules.print cps
        Rules.print cp
  | KH12 (rs, ss, ecp, proof) ->
      fprintf ppf 
        "---- CR(R cup S)@.\
         ---- [Klein and Hirokawa, 2012]@.\
         R:@.%a@.\
         S:@.%a@.\
         CP_S(R):@.%a@.\
         CR(S):@.see below.@.@.%a"
	Rules.print rs Rules.print ss
	Rules.print ecp print_proof proof
  | HM13 (rs, ss, cps, proof1, proof2) ->
      fprintf ppf 
        "---- CR(R cup S)@.\
         ---- [Hirokawa and Middeldorp, 2013]@.\
         R:@.%a@.\
         S:@.%a@.\
         CPS(R,S):@.%a@.\
         CR(R):@.see below.@.\
         CR(S):@.see below.@.@.%a%a"
	Rules.print rs Rules.print ss
	Rules.print cps 
        print_proof proof1
        print_proof proof2
  | HM13CR (rs, cps) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Hirokawa and Middeldorp, 2013]@.\
         R:@.%a@.\
         non-closed-CPS(R):@.%a"
	Rules.print rs
	Rules.print cps 
  | H80 (rs, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- R : linear and strongly closed [Huet, 1980]@.\
         R:@.%a@.\
         CP(R):@.%a"
	Rules.print rs
	Rules.print cp
  | HvONO18 (rs, cp, cs, proof) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- R : left-linear, C : CR and CP-closing, and Cdup/R : SN@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@.\
         critical-pair-closing system C:@.%a@.@.\
         CR(C): see below.@.@.%a@."
	Rules.print rs
	Rules.print cp
	Rules.print cs
	print_proof proof
  | HOT (rs, cs, a) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Hot decreasingness.@.\
         R:@.%a@.@.\
         C:@.%a@.@.\
         Label:@.%a@.@."
        Rules.print rs Rules.print cs (Listx.print print_label "@.") a
  | OH14b (rs, cp_out, cp_in, cs) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Oyamaguchi-Hirokawa, 2014]@.\
         R:@.%a@.@.\
         outside critical pairs of R:@.%a@.@.\
         inside critical pairs of R:@.%a@.@.\
         critical-pair-closing system C:@.%a@.@."
	Rules.print rs
	Rules.print cp_out
	Rules.print cp_in
	Rules.print cs
  | JK86 (rs, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Jouannaud and Kirchner, 1986]@.\
         R:@.%a@.\
         CP_AC(R):@.%a"
	Rules.print rs Rules.print cp
  | VO94 (rs, cp) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Development Closedness Theorem [van Oostrom, TCS 1994].@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@."
        Rules.print rs Rules.print cp
  | AYT09 (rs, cp) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Extended Development Closedness Theorem [Aoto, Yoshida and Toyama, RTA 2009].@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@."
        Rules.print rs Rules.print cp
  | VO08 (rs, a) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Rule Labeling [van Oostrom, RTA 2008].@.\
         R:@.%a@.@.\
         Label:@.%a@.@."
        Rules.print rs (Listx.print print_label "@.") a
  | Huet_conjecture -> fprintf ppf "Huet's conjecture@."

let print_disproof ppf (rs, s, t) =
  fprintf ppf
    "---- Non-confluence check@.\
     R:@.%a@.@.\
     Unjoinable conversion:@.%a <->*@.%a@.@." 
    Rules.print rs
    Term.print s Term.print t

(* Auxiliary functions *)

exception Failed
exception Disproved of Term.t * Term.t

let rec try_find f = function
  | [] -> raise Failed
  | x :: xs -> try f x with Failed -> try_find f xs

let rec pairwise_find p = function
  | [] -> raise Failed
  | x :: xs ->
      try 
        let y = List.find (p x) xs in (x, y) 
      with Not_found -> pairwise_find p xs

(* Global variables *)

let filenames  = ref []
let tt         = ref "natt -v:0"
let actt       = ref "muterm -i"
let smt        = ref "minismt -m"
let wcrk       = ref 5
let rlvk       = ref 5
let rlck       = ref 5

let rec tcap rs = function
  | V _ -> V (fresh ())
  | F (f, ts) -> 
      let u = F (f, [ tcap rs t | t <- ts ]) in 
      if List.exists (fun (l, _) -> unifiable u l) rs then
        V (fresh ())
      else
        u


let commute1 n rs ss (s, t) = 
  let ss = reachable n rs [s]
  and ts = reachable n ss [t] in
  Listx.intersect ss ts

let commute n rs ss es =
  List.for_all (commute1 n rs ss) es

let joinable n rs es = commute n rs rs es

(** Confluence criteria *)

(* Knuth and Bendix' criterion (1970) *)
let kb70 rs = 
  let cp = cp rs in
  if not (Rsn.not_sn rs []) && 
     Sn.sn !tt rs [] = YES &&
     List.for_all (fun (s, t) -> nf rs s = nf rs t) cp
  then
    (*KH12 (rs, [], cp, HM11 ([], [], []))*)
    KB70 (rs, cp)
  else
    raise Failed

(* Hirokawa and Middeldorp's criterion (JAR 2011) *)
let hm11 rs = 
  let p (s, t) = s <> t in
  let cps = cps ~p rs in
  let cp = cp rs in
  if Rules.left_linear rs && 
     not (Rsn.not_sn cps rs) && 
     joinable !wcrk rs cp &&
     Sn.sn !tt cps rs = YES
  then
    HM11 (rs, cps, cp)
  else 
    raise Failed

(* Development closedness theorem (Van Oostrom, TCS 1994) *)
let vo94 rs = 
  if Rules.left_linear rs then 
     let p (s, t) = List.mem t (multistep rs s) in
     let cp = cp rs in
     if List.for_all p cp then
       VO94 (rs, cp)
     else 
       raise Failed
  else
    raise Failed

(* Extended development closedness theorem (Aoto, Yoshida and Toyama, RTA 2009) *)
let root_overlap (_, p, _, _) = p = []

let closed rs o = 
  let s, t = cp_of_overlap o in
  if root_overlap o then 
    Listx.intersect (multistep rs s) (reachable !wcrk rs [t])
  else
    List.mem t (multistep rs s)

let ayt09 rs = 
  if Rules.left_linear rs &&
     List.for_all (closed rs) (overlap rs) then 
    AYT09 (rs, cp rs)
  else
    raise Failed

(* Van Oostrom's rule labeling (RTA 2008) *)

let rlv rs = 
  match Rlv.solve ~tool:!smt !rlvk rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

let rlc rs =
  match Rl.solve  ~tool:!smt !rlck rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

(* Non-confluence check *)

let unjoinable_by_tcap rs s t = 
  not (Substitution.unifiable (tcap rs s) (tcap rs t))

(*
let unjoinable_by_closure k rs s t = 
  let ss1 = reachable k rs [s]
  and ts1 = reachable k rs [t] in
  let ss2 = reachable 1 rs ss1
  and ts2 = reachable 1 rs ts1 in
  Listx.subset ss2 ss1 &&
  Listx.subset ts2 ts1 && 
  not (intersect ss2 ts2)

let unjoinable k rs s t =
  unjoinable_by_tcap rs s t ||
  unjoinable_by_closure 1 rs s t
*)

let rec ground = function
  | V x -> F (x, [])
  | F (f, ts) -> F (f, [ ground t | t <- ts ])

let critical_peak rs =
  [ ground (substitute mu l2) | (_, _), _, (l2, _), mu <- overlap rs ]

let non_confluent k rs =
  let f t = pairwise_find (unjoinable_by_tcap rs) (reachable k rs [t]) in
  try_find f (critical_peak (rs @ [ r, l | l, r <- rs]))

(* Klein and Hirokawa's criterion (LPAR 2012) *)

(* joinability of S-overlap *)

let kh12_overlap_aux rule1 rule2 =
  let l1, r1 = Rule.rename rule1
  and l2, r2 = Rule.rename rule2 in
  [ (l1, r1), p, (l2, r2), mgu (subterm_at p l2) l1
  | p <- function_positions l2; 
    unifiable (ren l1) (ren (subterm_at p l2));
    p <> [] || not (Rule.variant (l1, r1) (l2, r2)) ] 

let kh12_overlap rs =
  Listx.unique [ o | a <- rs; b <- rs; o <- kh12_overlap_aux a b ]

let kh12_cp rs = [ cp_of_overlap o | o <- kh12_overlap rs ]

let kh12 cr rs ss =
  let ecp =
    try kh12_cp rs with Not_unifiable -> raise Failed in
  if sno rs ss && 
     joinable !wcrk (rs @ ss) ecp &&
     Sn.sn !tt rs ss = YES
  then
    KH12 (rs, ss, ecp, cr ss)
  else
    raise Failed

let kh12_auto cr rs =
  let f ss = kh12 cr (Listx.diff rs ss) ss in
  try_find f [ ss | ss <- Listx.power rs; ss <> [] && not (Listx.subset rs ss) ]

(* Jouannaud and Kirchner's criterion (1986) *)

let jk86 rr =
  match Ac_rewriting.find_ac rr with
  | None -> raise Failed
  | Some (rs, ac_sym, ac, ac_all) ->
      if not (Sn.sn !actt ~ac_sym rs [] = YES) then 
	raise Failed
      else 
	let cps = Rules.flatten ac_sym
          (Ac_overlap.cp  ac_sym rs @
           Ac_overlap.cp2 ac_sym rs ac_all) in
	let cps_nf =
           [ Ac_rewriting.nf ac_sym rs s,
             Ac_rewriting.nf ac_sym rs t 
           | s, t <- cps ] in
	let nj =
          [ s,t | s,t <- Variant.rename_rules cps_nf; 
                  not (Ac_rewriting.equal ac_sym s t) ] in
        match nj with
        | [] -> JK86 (rr, Variant.rename_rules cps)
        | (s, t) :: _ -> raise (Disproved (s, t))

let jk86_maybe rr = 
  try jk86 rr with Disproved (_, _) -> raise Failed

(* Hirokawa and Middeldorp's criterion (2013) *)

(*
let hm13 cr rs ss =
  let p (s, t) = true (* s <> t *) in
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = mutual_cps ~p rs ss in
     if Sn.sn !tt cps union = YES then
       HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13 cr rs1 rs2 in
    let l = Listx.power rs in
    try_find f
      [ rs1, rs2
      | rs1 <- l; rs2 <- l;
        not (Listx.equal rs1 rs2);
        Listx.equal rs (rs1 @ rs2) ]
  else
    raise Failed

let hm13b_non_closed_overlap rs (s, t) =
  not (List.mem t (multistep rs s))

let hm13b_cps rs ss =
  Listx.unique
    (cps2 ~p:(hm13b_non_closed_overlap ss) rs ss @
     cps2 ~p:(hm13b_non_closed_overlap rs) ss rs)

let hm13b cr rs ss =
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = hm13b_cps rs ss in
     if Sn.sn !tt cps union = YES then
       if Listx.equal rs ss then
         HM13CR (rs, cps)
       else
         HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13b_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13b cr rs1 rs2 in
    let l = Listx.power rs in
    try_find f
      [ rs1, rs2
      | rs1 <- l; rs2 <- l;
        not (Listx.equal rs1 rs2);
        Listx.equal rs (rs1 @ rs2) ]
  else
    raise Failed

let hm13b_cr rs =
  let cr rs = raise Failed in
  hm13b cr rs rs
*)

let hm13c_cps rs ss =
  let peaks =
    [ peak | o <- overlap2 rs ss; 
	     (t, s, u) as peak <- [critical_peak_of_overlap o];
	     not (Listx.intersect (multistep ss t) (reachable !wcrk rs [u])) ] @
    [ peak | o <- overlap2 ss rs; not (root_overlap o);
	     (t, s, u) as peak <- [critical_peak_of_overlap o];
	     not (List.mem u (multistep rs t)) ] in
  Listx.unique [ rule | t, s, u <- peaks; rule <- [s, t; s, u] ]

let hm13c cr rs ss =
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = hm13c_cps rs ss in
     if Sn.sn !tt cps union = YES then
       if Listx.equal rs ss then
         HM13CR (rs, cps)
       else
         HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13c_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13c cr rs1 rs2 in
    let l = Listx.power rs in
    try_find f
      [ rs1, rs2
      | rs1 <- l; rs2 <- l;
        not (Listx.equal rs1 rs2);
        Listx.equal rs (rs1 @ rs2) ]
  else
    raise Failed

let hm13 rs =
  let cr rs = raise Failed in
  hm13c cr rs rs

(* Emptiness *)

let emptiness = function
  | [] -> EMPTY
  | _  -> raise Failed

(* Strong Closedness Theorem, Huet 1980 *)

let strongly_closed1 rs (s, t) =
  Listx.intersect (reachable 1 rs [s]) (reachable !wcrk rs [t])

let strong_closedness rs =
  if Rules.linear rs then
    let cps = cp rs in
    let l = Listx.unique [ x | (s, t) <- cps; x <- [(s, t); (t, s)] ] in
    if List.for_all (strongly_closed1 rs) l then
      H80 (rs, cps)
    else
      raise Failed
  else
    raise Failed

(* Generalized Strong Closedness Theorem *)

let generalized_strong_closedness rs =
  if Rules.left_linear rs then
    let cps = cp rs in
    let critical_pair_closing cs =
      List.for_all (Rewriting.joinable !wcrk cs) cps in
    let f cs = HvONO18 (rs, cps, cs, strong_closedness cs) in
    try_find f [ cs | cs <- Listx.power rs; critical_pair_closing cs ]
  else
    raise Failed

(* HvONO 2018 *)

let hvono18 cr rs cs =
  let cs_d = [ rule | rule <- cs; not (Rule.is_non_duplicating rule) ] in
  if Rules.left_linear rs && Sn.sn !tt cs_d rs = YES then
    let cps = cp rs in
    if cps = [] then
      HvONO18 (rs, [], [], cr [])
    else if List.for_all (Rewriting.joinable !wcrk cs) cps then
      HvONO18 (rs, cps, cs, cr cs)
    else
      raise Failed
  else
   raise Failed

let hvono18_auto cr rs =
  if Rules.left_linear rs then
    try_find (hvono18 cr rs)
      [ cs | cs <- Listx.power rs; not (Listx.equal rs cs) ]
  else
    raise Failed

let rec hvono18a rs =
  try_find (fun f -> f rs) [emptiness; hvono18_auto hvono18a]

let rec hvono18b rs =
  try_find (fun f -> f rs) [strong_closedness; hvono18_auto hvono18b]

(* Hot decreasingness *)
    
let hot rs =
  let f cs = 
    match Hot.hot ~yices:"yices" ~termination_tool:!tt !wcrk rs cs with
    | Some a -> HOT (rs, cs, a)
    | None -> raise Failed in
  try_find f (Listx.power rs)
  

    
(* OH14b *)

let oh14b_outside_closed cs (s, t) = nf cs s = nf cs t

let oh14b_inside_closed cs (s, t) =
  Listx.intersect (multistep cs s) (reachable_all cs [t])

let is_overlay (_,p,_,_) = p = []

let cp_out_in rs =
  let os_out, os_in = List.partition is_overlay (Overlap.overlap rs) in
  ([ Overlap.cp_of_overlap o | o <- os_out ],
   [ Overlap.cp_of_overlap o | o <- os_in ])

let oh14b rs cs =
  let cp_out, cp_in = cp_out_in rs in
  if Rules.left_linear rs &&
     (* Sn.sn !tt cs [] = YES && *)
     List.for_all (oh14b_outside_closed cs) cp_out &&
     List.for_all (oh14b_inside_closed cs) cp_in 
  then
    OH14b (rs, cp_out, cp_in, cs)
  else
    raise Failed

let oh14b_auto rs =
  if Rules.left_linear rs then
    try_find (oh14b rs) (Cpc.cpc rs)
  else
    raise Failed

let parallel_either rs (s, t) =
  List.mem t (parallel_step rs s) ||
  List.mem s (parallel_step rs t)

let huet_conjecture rs =
  if Rules.left_linear rs && 
     List.for_all (parallel_either rs) (cp rs)
  then
    Huet_conjecture
  else
    raise Failed

let rec confluent1 rs = 
  try_find (fun f -> f rs)
    [ kb70; hm13; rlv; rlc; (*jk86_maybe; *) kh12_auto confluent1]

let rec confluent rs = 
  try_find (fun f -> f rs)
    [ kb70; strong_closedness; vo94; 
      hm13; rlv; rlc; hvono18_auto confluent; kh12_auto confluent]

(*
let confluent rs = huet_conjecture rs
*)

(*
let rec confluent rs = 
  try_find (fun f -> f rs) [oh14b_auto; oh14a_auto confluent]
*)

(*
module M = Map.Make
  (struct
     type t = Rules.t
     let compare rs ss =
       Pervasives.compare
	 (List.sort Pervasives.compare rs)
	 (List.sort Pervasives.compare ss)
   end)

let memo = ref M.empty

let rec confluent rs = 
  try 
    M.find rs !memo
  with Not_found ->
    let b =
      try_find (fun f -> f rs)
	[ hm13c_cr; rlv; hm13c_auto confluent ] in
    (memo := M.add rs b !memo; b)
*)

let prove = ref confluent

let options = align [ 
  "-kb", Unit (fun () -> prove := kb70),
     sprintf " Knuth and Bendix' criterion (1970)";
  "-hm", Unit (fun () -> prove := hm11),
     sprintf " critical pair system (Hirokawa and Middeldorp 2011)";
  "-rlv", Unit (fun () -> prove := rlv),
     sprintf " valley version of rule labeling (van Ootrom 2008)";
  "-dc", Unit (fun () -> prove := vo94),
     sprintf " development closeness theorem (van Ootrom 1980)";
  "-sc", Unit (fun () -> prove := strong_closedness),
     sprintf " strong closeness theorem (Huet 1980)";
  "-gsc", Unit (fun () -> prove := generalized_strong_closedness),
     sprintf " generalized strong closeness theorem";
  "-cps", Unit (fun () -> prove := hvono18a),
     sprintf " critical-pair-closing + emptiness";
  "-hot", Unit (fun () -> prove := hot),
     sprintf " hot decreasingness";
  "-oh14b", Unit (fun () -> prove := oh14b_auto),
     sprintf " OH14b";
  "-tt", String (fun x -> tt := x),
     sprintf "<tool> termination tool (default: %s)" !tt;
  "-actt", String (fun x -> actt := x),
     sprintf "<tool> AC termination tool (default: %s)" !actt;
  "-smt", String (fun x -> smt := x),
     sprintf "<tool> QF_NIA SMT solver (default: %s)" !smt;
  "-k",   Int (fun n -> wcrk := n),
     sprintf "<n> at most k steps for joinability test (default: %d)" !wcrk;
]

(* main *)
     
let short_usage = "Usage: saigawa [options] <file>\n"
let usage = "Confluence tool Saigawa v1.10\n" ^ short_usage ^ "Options are:"

let () = 
  parse options (fun it -> filenames := !filenames @ [it] ) usage;
  if List.length !filenames <> 1 then
    (eprintf "%s%!" short_usage; exit 1)
  else
    let filename = List.hd !filenames in
    let rs = Read.read_trs filename in
    if not (Rules.variable_condition rs) then
      (eprintf "Cannot handle TRS with extra variables.@."; exit 1);
(*
    try 
      let s, t = non_confluent !wcrk rs in
      printf "NO@.@.%a@." print_disproof (rs, s, t)
    with Failed ->
*)
      try 
	let proof = !prove rs in
	printf "YES@.@.%a@." print_proof proof
      with
      | Disproved (s, t) -> printf "NO@.@.%a@." print_disproof (rs, s, t)
      | Failed -> printf "MAYBE@."
