(** Yices *)

open Format

type 'a assignment =
  { int : ('a * int) list;
    num : ('a * Num.num) list;
    bool : ('a * bool) list } 

type 'a t =
  | BVar of 'a
  | Not of 'a t
  | And of 'a t list
  | Or of 'a t list
  | Implies of 'a t * 'a t
  | Eqv of 'a t * 'a t
  | Eq  of 'a e * 'a e
  | Geq of 'a e * 'a e
  | Gt  of 'a e * 'a e
and 'a e = 
  | IVar of 'a 
  | Int of int
  | Plus of 'a e list
  | Ite of 'a t * 'a e * 'a e

let debug = ref false

let print_args ppf op e pr = function
  | []  -> fprintf ppf "%s" e
  | [x] -> pr ppf x
  | xs  -> fprintf ppf "(%s %a)" op (Listx.print pr " ") xs

let rec print_e to_s ppf = function
  | Int n   -> fprintf ppf "%d" n
  | IVar x  -> fprintf ppf "%s" (to_s x)
  | Plus xs -> print_args ppf "+" "0" (print_e to_s) xs
  | Ite (c, e1, e2) ->
      fprintf ppf "(ite %a %a %a)"
        (print_c to_s) c (print_e to_s) e1 (print_e to_s) e2

and print_c to_s ppf = function
  | BVar x  -> fprintf ppf "%s" (to_s x)
  | Eqv (c1, c2) ->
      fprintf ppf "(= %a %a)"  (print_c to_s) c1 (print_c to_s) c2
  | Eq  (e1, e2) ->
      fprintf ppf "(= %a %a)"  (print_e to_s) e1 (print_e to_s) e2
  | Gt  (e1, e2) ->
      fprintf ppf "(> %a %a)"  (print_e to_s) e1 (print_e to_s) e2
  | Geq (e1, e2) ->
      fprintf ppf "(>= %a %a)" (print_e to_s) e1 (print_e to_s) e2
  | Not c  -> fprintf ppf "(not %a)" (print_c to_s) c
  | And cs -> print_args ppf "and" "true"  (print_c to_s) cs
  | Or  cs -> print_args ppf "or"  "false" (print_c to_s) cs
  | Implies (c1, c2) ->
      fprintf ppf "(=> %a %a)" (print_c to_s) c1 (print_c to_s) c2

let neg = function
  | And [] -> Or []
  | Or []  -> And []
  | c      -> Not c

let implies x y = 
  match x, y with
  | Or [], _ 
  | _, And [] -> And []
  | And [], y -> y
  | x, Or []  -> neg x
  | _         -> Implies (x, y)

let disj l = Or l

let conj l = And l

let geq x y = Geq (x, y)

let gt x y = Gt (x, y)

let plus = function
  | []  -> Int 0
  | [x] -> x
  | l   -> Plus l

let ite b t e =
  match b with
  | And [] -> t
  | Or []  -> e 
  | _      -> Ite (b, t, e)


(* A parser for outputs *)

let finally x f g =
  let r = try f x with e -> g x; raise e in 
  g x; r

let read_channel ?(filename = "") ch =
  let lb0 = Lexing.from_channel ch in
  let lex_curr_p = { lb0.Lexing.lex_curr_p with Lexing.pos_fname = filename } in
  let lb = { lb0 with Lexing.lex_curr_p } in
  try
    Yices_parser.yices Yices_lexer.token lb
  with
    | Parsing.Parse_error ->
        eprintf "%a%!" Syntaxerr.print (lb, lb.Lexing.lex_curr_p, "Parse error.");
        exit 1
    | Syntaxerr.Syntax_error (p, msg) -> 
        eprintf "%a%!" Syntaxerr.print (lb, p, msg);
        exit 1

let assoc e x a = 
  try List.assoc x a with Not_found -> e

let read_yices_from_channel ~int ~num ~bool ~to_string ch =
  match finally ch read_channel close_in with
  | "unsat", _ -> None
  | "sat",   l ->
      let a = [ x, s | ["="; x; s] <- l ] in
      Some
        { int  = [ x, int_of_string     (assoc "0" (to_string x) a)     | x <- int  ];
          num  = [ x, Num.num_of_string (assoc "0" (to_string x) a)     | x <- num  ];
          bool = [ x, bool_of_string    (assoc "false" (to_string x) a) | x <- bool ] }
  | _ -> failwith "Read.read_yices_from_channel"

let print_define to_s ppf (x, t) = 
  fprintf ppf "(define %s :: %s)@." (to_s x) t

let print_define_list to_s ppf (xs, t) =
  List.iter (print_define to_s ppf) [ (x, t) | x <- Listx.unique xs ]

let print to_s ppf (int, num, bool, c) =
  fprintf ppf "%a%a%a(assert %a)@.(check)@.(show-model)@."
    (print_define_list to_s) (int,  "int")
    (print_define_list to_s) (num,  "real")
    (print_define_list to_s) (bool, "bool")
    (print_c to_s) c

let print_result ppf = function
  | None -> fprintf ppf "unsat@."
  | Some l ->
      fprintf ppf "sat@.";
      List.iter
        (fun (x, y) -> fprintf ppf "%s = %s@." x (Num.string_of_num y)) l

let sat ?(tool = "yices") 
        ?(int = []) ?(num = []) ?(bool = []) 
        ~to_string c =
  let s = Filename.temp_file "" ".ys" in
  finally 
    (open_out s) 
    (fun ch ->
      if !debug then print to_string err_formatter (int, num, bool, c);
      print to_string (formatter_of_out_channel ch) (int, num, bool, c))
    close_out;
  finally 
    (Unix.open_process_in (sprintf "%s %s 2> /dev/null" tool s))
    (read_yices_from_channel ~int ~num ~bool ~to_string)
    (fun ch -> ignore (Unix.close_process_in ch); Sys.remove s)
